"""   
Minimal example of truss optimization in 2D

This script computes the optimal choice of bars (pairwise undirected connection of nodes)
for a given set of nodes, representing either a discrete structure or a discretized continuum.
By optimal we mean the minimum of material (volume), 
where the bar length is determined by the spatial distance between two nodes
and the cross section area by the axial force and the material yield stress.
The axial forces in the bars are the optimization variables.
They are subject to the constraint of mechanical equilibrium at each node. 
The material distribution (cross section area) follows from the axial forces 
by the assumption, that the yield stress is reached in each bar (ideal-plastic).
Be aware, that buckling is ignored, but must be accounted for in practical applications!
For computational efficiency the axial force is split into two values: tension and compression,
both constrained to non-negativity.

As example, we consider a bridge-like configuration of a rectangular domain
with two fixed nodes on left and right end of its top boundary and 
a vertical point load applied at the central node of the top boundary.
For sake of simplicity, we assume a unit load and a unit yield stress.
The node numbers for boundary conditions and load are hard-coded, 
but their choice can be comprehend by the plot initially commented out.

For the theory behind and a more advanced example see 'A Python script for adaptive layout optimization of trusses' by Linwei He, Matthew Gilbert and Xingyi Song.
https://doi.org/10.1007/s00158-019-02226-6
"""

import cvxpy as cp
import numpy as np
import matplotlib.pyplot as plt

"""   nodal discretization   """
Nx, Ny = (5, 3)
x = np.linspace(0, 1, Nx)
y = np.linspace(0, 1, Ny)
xv, yv = np.meshgrid(x, y)
Nnode = Nx*Ny
node = np.zeros((Nnode,2))
node[:,0] = xv.flatten()
node[:,1] = yv.flatten()

"""   show nodes to choose node numbers for B.C. and load   """
# for n,onenode in enumerate(node):
#     plt.plot(onenode[0], onenode[1], 'k.')
#     plt.annotate(str(n),(onenode[0], onenode[1]))
# plt.title('Node numbering')
# plt.axis('off')
# plt.show()

"""   B.C. and load   """
f = np.zeros(2*Nnode)    # external nodal forces with two components (x,y)
n_fix =  np.array([10, 14])  # fixed nodes (ux=uy=0 at left and right end of top boundary)
n_load = np.array([12])      # loaded node (center of top boundary)
FX =  0                     
FY = -1 # downward force 
f[2*n_load]   = FX
f[2*n_load+1] = FY   

"""   optimization variable: bars   """
bar = [(i, j) for i in range(Nnode) for j in range(Nnode) if j > i]   # undirected connection of two nodes
Nbar = len(bar)
print("{} nodes and {} bars".format(Nnode, Nbar))

c = np.zeros(2*Nbar)   # volume factor (bar length divided by yield stress)
B = np.zeros((2*Nnode, 2*Nbar))   # (mechanical) equilibrium matrix
for m,onebar in enumerate(bar):
    n0 = onebar[0]  # first node
    n1 = onebar[1]  # second node
    dx = node[n1, 0] - node[n0, 0]
    dy = node[n1, 1] - node[n0, 1]
    l  = np.sqrt(dx**2 + dy**2)   # lenght of bar
    n0x = 2*n0      # row index for x-contribution of node n0
    n0y = 2*n0+1    # row index for y-contribution of node n0
    n1x = 2*n1      # row index for x-contribution of node n1
    n1y = 2*n1+1    # row index for y-contribution of node n1
    
    mp = 2*m   # tension 
    c[mp] = l       # volume factor assuming unit yield stress (sigma_y=1)
    B[n0x, mp]   =  dx/l
    B[n0y, mp]   =  dy/l
    B[n1x, mp]   = -dx/l
    B[n1y, mp]   = -dy/l
    
    mm = 2*m+1   # compression
    c[mm] = l
    B[n0x, mm] = -dx/l
    B[n0y, mm] = -dy/l
    B[n1x, mm] =  dx/l
    B[n1y, mm] =  dy/l

# evaluate BC
f_red = np.delete(f, 2*n_fix)		# fx
f_red = np.delete(f, 2*n_fix+1)	# fy
B_red = np.delete(B, 2*n_fix, 0)	# ux=0
B_red = np.delete(B, 2*n_fix+1, 0)	# uy=0

"""   solve optimization problem numerically   """
q = cp.Variable(2*Nbar)   # axial forces in bars split into tension and compression
cost = cp.Minimize(c.transpose() @ q)   # material volume as cost function
cons = []
cons.append(B_red @ q + f_red == 0)  # equilibrium
cons.append(q >= 0)  # non-negativity
prob = cp.Problem(cost, cons)
optimal_design = prob.solve()

"""   plot results   """
scale_force_for_plot = 0.2   # TODO detect automatically to fit in spatial dimensions       
for n,onenode in enumerate(node):
    if n in n_fix:
        plt.plot(onenode[0], onenode[1], 'kx')
    else:
        plt.plot(onenode[0], onenode[1], 'k.')
    if n in n_load:
        DX = scale_force_for_plot*FX
        DY = scale_force_for_plot*FY
        plt.arrow(onenode[0], onenode[1], DX, DY, color='red', head_starts_at_zero=False, head_width=0.02)        

qmax = np.max(q.value) 
qmin = np.min(q.value) 
if qmin < 0:   # normalized value [0...1] to make opacity proportional to force
    qplot = (q.value-qmin)/(qmax-qmin)   
else:
    qplot = q.value/qmax    
    
for m,onebar in enumerate(bar):
    n0 = onebar[0]
    n1 = onebar[1]
    plt.plot( [node[n0,0], node[n1,0]], [node[n0,1], node[n1,1]], 'b', alpha=qplot[2*m])   
    plt.plot( [node[n0,0], node[n1,0]], [node[n0,1], node[n1,1]], 'g', alpha=qplot[2*m+1]) 

plt.plot( [0,0], [0,0], 'b', label='tension')   # fake line for legend independent on opacity
plt.plot( [0,0], [0,0], 'g', label='compression')   # fake line for legend independent on opacity   
plt.legend(loc='lower right')
plt.axis('off')
plt.title('optimal truss')
plt.show()
